@@ -1,4 +1,4 @@
-// Copyright 2019 Parity Technologies (UK) Ltd.
+// Copyright 2019-2021 Parity Technologies (UK) Ltd.
 //
 // Permission is hereby granted, free of charge, to any
 // person obtaining a copy of this software and associated
@@ -24,15 +24,31 @@
 // IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 // DEALINGS IN THE SOFTWARE.
 
-use crate::tokio::{TcpStream, TlsStream};
+use crate::stream::EitherStream;
+use arrayvec::ArrayVec;
 use futures::io::{BufReader, BufWriter};
-use futures::prelude::*;
+use http::Uri;
 use soketto::connection;
-use soketto::handshake::client::{Client as WsRawClient, ServerResponse};
-use std::{borrow::Cow, io, net::SocketAddr, sync::Arc, time::Duration};
+use soketto::handshake::client::{Client as WsHandshakeClient, Header, ServerResponse};
+use std::convert::TryInto;
+use std::{
+	borrow::Cow,
+	convert::TryFrom,
+	io,
+	net::{SocketAddr, ToSocketAddrs},
+	sync::Arc,
+	time::Duration,
+};
 use thiserror::Error;
+use tokio::net::TcpStream;
+use tokio_rustls::{
+	client::TlsStream,
+	rustls::ClientConfig,
+	webpki::{DNSNameRef, InvalidDNSNameError},
+	TlsConnector,
+};
 
-type TlsOrPlain = crate::stream::EitherStream<TcpStream, TlsStream<TcpStream>>;
+type TlsOrPlain = EitherStream<TcpStream, TlsStream<TcpStream>>;
 
 /// Sending end of WebSocket transport.
 #[derive(Debug)]
@@ -60,6 +76,8 @@ pub struct WsTransportClientBuilder<'a> {
 	pub origin_header: Option<Cow<'a, str>>,
 	/// Max payload size
 	pub max_request_body_size: u32,
+	/// Max number of redirections.
+	pub max_redirections: usize,
 }
 
 /// Stream mode, either plain TCP or TLS.
@@ -88,42 +106,42 @@ pub enum CertificateStore {
 #[derive(Debug, Error)]
 pub enum WsHandshakeError {
 	/// Failed to load system certs
-	#[error("Failed to load system certs: {}", 0)]
+	#[error("Failed to load system certs: {0}")]
 	CertificateStore(io::Error),
 
 	/// Invalid URL.
-	#[error("Invalid url: {}", 0)]
+	#[error("Invalid URL: {0}")]
 	Url(Cow<'static, str>),
 
 	/// Error when opening the TCP socket.
-	#[error("Error when opening the TCP socket: {}", 0)]
+	#[error("Error when opening the TCP socket: {0}")]
 	Io(io::Error),
 
 	/// Error in the transport layer.
-	#[error("Error in the WebSocket handshake: {}", 0)]
+	#[error("Error in the WebSocket handshake: {0}")]
 	Transport(#[source] soketto::handshake::Error),
 
 	/// Invalid DNS name error for TLS
-	#[error("Invalid DNS name: {}", 0)]
-	InvalidDnsName(#[source] crate::tokio::InvalidDNSNameError),
+	#[error("Invalid DNS name: {0}")]
+	InvalidDnsName(#[source] InvalidDNSNameError),
 
-	/// RawServer rejected our handshake.
-	#[error("Connection rejected with status code: {}", status_code)]
+	/// Server rejected the handshake.
+	#[error("Connection rejected with status code: {status_code}")]
 	Rejected {
 		/// HTTP status code that the server returned.
 		status_code: u16,
 	},
 
 	/// Timeout while trying to connect.
-	#[error("Connection timeout exceeded: {}", 0)]
+	#[error("Connection timeout exceeded: {0:?}")]
 	Timeout(Duration),
 
 	/// Failed to resolve IP addresses for this hostname.
-	#[error("Failed to resolve IP addresses for this hostname: {}", 0)]
+	#[error("Failed to resolve IP addresses for this hostname: {0}")]
 	ResolutionFailed(io::Error),
 
 	/// Couldn't find any IP address for this hostname.
-	#[error("No IP address found for this hostname: {}", 0)]
+	#[error("No IP address found for this hostname: {0}")]
 	NoAddressFound(String),
 }
 
@@ -148,6 +166,11 @@ impl Sender {
 		self.inner.flush().await?;
 		Ok(())
 	}
+
+	/// Send a close message and close the connection.
+	pub async fn close(&mut self) -> Result<(), WsError> {
+		self.inner.close().await.map_err(Into::into)
+	}
 }
 
 impl Receiver {
@@ -164,7 +187,7 @@ impl<'a> WsTransportClientBuilder<'a> {
 	pub async fn build(self) -> Result<(Sender, Receiver), WsHandshakeError> {
 		let connector = match self.target.mode {
 			Mode::Tls => {
-				let mut client_config = rustls::ClientConfig::default();
+				let mut client_config = ClientConfig::default();
 				if let CertificateStore::Native = self.certificate_store {
 					client_config.root_store = rustls_native_certs::load_native_certs()
 						.map_err(|(_, e)| WsHandshakeError::CertificateStore(e))?;
@@ -174,74 +197,143 @@ impl<'a> WsTransportClientBuilder<'a> {
 			Mode::Plain => None,
 		};
 
-		let mut err = None;
-		for sockaddr in &self.target.sockaddrs {
-			match self.try_connect(*sockaddr, &connector).await {
-				Ok(res) => return Ok(res),
-				Err(e) => {
-					log::debug!("Failed to connect to sockaddr: {:?} with err: {:?}", sockaddr, e);
-					err = Some(Err(e));
-				}
-			}
-		}
-		// NOTE(niklasad1): this is most likely unreachable because [`Url::socket_addrs`] doesn't
-		// return an empty `Vec` if no socket address was found for the host name.
-		err.unwrap_or(Err(WsHandshakeError::NoAddressFound(self.target.host)))
+		self.try_connect(connector).await
 	}
 
 	async fn try_connect(
-		&self,
-		sockaddr: SocketAddr,
-		tls_connector: &Option<crate::tokio::TlsConnector>,
+		self,
+		mut tls_connector: Option<TlsConnector>,
 	) -> Result<(Sender, Receiver), WsHandshakeError> {
-		// Try establish the TCP connection.
-		let tcp_stream = {
-			let socket = TcpStream::connect(sockaddr);
-			let timeout = crate::tokio::sleep(self.timeout);
-			futures::pin_mut!(socket, timeout);
-			match future::select(socket, timeout).await {
-				future::Either::Left((socket, _)) => {
-					let socket = socket?;
-					if let Err(err) = socket.set_nodelay(true) {
-						log::warn!("set nodelay failed: {:?}", err);
-					}
-					match tls_connector {
-						None => TlsOrPlain::Plain(socket),
-						Some(connector) => {
-							let dns_name = crate::tokio::DNSNameRef::try_from_ascii_str(&self.target.host)?;
-							let tls_stream = connector.connect(dns_name, socket).await?;
-							TlsOrPlain::Tls(tls_stream)
-						}
-					}
-				}
-				future::Either::Right((_, _)) => return Err(WsHandshakeError::Timeout(self.timeout)),
-			}
-		};
+		let mut target = self.target;
+		let mut headers: ArrayVec<Header, 1> = ArrayVec::new();
+		let mut err = None;
 
-		log::debug!("Connecting to target: {:?}", self.target);
-		let mut client = WsRawClient::new(
-			BufReader::new(BufWriter::new(tcp_stream)),
-			&self.target.host_header,
-			&self.target.path_and_query,
-		);
 		if let Some(origin) = self.origin_header.as_ref() {
-			client.set_origin(origin);
+			headers.push(Header { name: "Origin", value: origin.as_bytes() });
 		}
 
-		// Perform the initial handshake.
-		match client.handshake().await? {
-			ServerResponse::Accepted { .. } => {}
-			ServerResponse::Rejected { status_code } | ServerResponse::Redirect { status_code, .. } => {
-				// TODO: HTTP redirects also lead here #339.
-				return Err(WsHandshakeError::Rejected { status_code });
+		for _ in 0..self.max_redirections {
+			log::debug!("Connecting to target: {:?}", target);
+
+			// The sockaddrs might get reused if the server replies with a relative URI.
+			let sockaddrs = std::mem::take(&mut target.sockaddrs);
+			for sockaddr in &sockaddrs {
+				let tcp_stream = match connect(*sockaddr, self.timeout, &target.host, &tls_connector).await {
+					Ok(stream) => stream,
+					Err(e) => {
+						log::debug!("Failed to connect to sockaddr: {:?}", sockaddr);
+						err = Some(Err(e));
+						continue;
+					}
+				};
+				let mut client = WsHandshakeClient::new(
+					BufReader::new(BufWriter::new(tcp_stream)),
+					&target.host_header,
+					&target.path_and_query,
+				);
+
+				client.set_headers(&headers);
+
+				// Perform the initial handshake.
+				match client.handshake().await {
+					Ok(ServerResponse::Accepted { .. }) => {
+						log::info!("Connection established to target: {:?}", target);
+						let mut builder = client.into_builder();
+						builder.set_max_message_size(self.max_request_body_size as usize);
+						let (sender, receiver) = builder.finish();
+						return Ok((Sender { inner: sender }, Receiver { inner: receiver }));
+					}
+
+					Ok(ServerResponse::Rejected { status_code }) => {
+						log::debug!("Connection rejected: {:?}", status_code);
+						err = Some(Err(WsHandshakeError::Rejected { status_code }));
+					}
+					Ok(ServerResponse::Redirect { status_code, location }) => {
+						log::debug!("Redirection: status_code: {}, location: {}", status_code, location);
+						match location.parse::<Uri>() {
+							// redirection with absolute path => need to lookup.
+							Ok(uri) => {
+								// Absolute URI.
+								if uri.scheme().is_some() {
+									target = uri.try_into()?;
+									tls_connector = match target.mode {
+										Mode::Tls => {
+											let mut client_config = ClientConfig::default();
+											if let CertificateStore::Native = self.certificate_store {
+												client_config.root_store = rustls_native_certs::load_native_certs()
+													.map_err(|(_, e)| WsHandshakeError::CertificateStore(e))?;
+											}
+											Some(Arc::new(client_config).into())
+										}
+										Mode::Plain => None,
+									};
+									break;
+								}
+								// Relative URI.
+								else {
+									// Replace the entire path_and_query if `location` starts with `/` or `//`.
+									if location.starts_with('/') {
+										target.path_and_query = location;
+									} else {
+										match target.path_and_query.rfind('/') {
+											Some(offset) => {
+												target.path_and_query.replace_range(offset + 1.., &location)
+											}
+											None => {
+												err = Some(Err(WsHandshakeError::Url(
+													format!(
+														"path_and_query: {}; this is a bug it must contain `/` please open issue",
+														location
+													)
+													.into(),
+												)));
+												continue;
+											}
+										};
+									}
+									target.sockaddrs = sockaddrs;
+									break;
+								}
+							}
+							Err(e) => {
+								err = Some(Err(WsHandshakeError::Url(e.to_string().into())));
+							}
+						};
+					}
+					Err(e) => {
+						err = Some(Err(e.into()));
+					}
+				};
 			}
 		}
+		err.unwrap_or(Err(WsHandshakeError::NoAddressFound(target.host)))
+	}
+}
 
-		// If the handshake succeeded, return.
-		let mut builder = client.into_builder();
-		builder.set_max_message_size(self.max_request_body_size as usize);
-		let (sender, receiver) = builder.finish();
-		Ok((Sender { inner: sender }, Receiver { inner: receiver }))
+async fn connect(
+	sockaddr: SocketAddr,
+	timeout_dur: Duration,
+	host: &str,
+	tls_connector: &Option<TlsConnector>,
+) -> Result<EitherStream<TcpStream, TlsStream<TcpStream>>, WsHandshakeError> {
+	let socket = TcpStream::connect(sockaddr);
+	let timeout = tokio::time::sleep(timeout_dur);
+	tokio::select! {
+		socket = socket => {
+			let socket = socket?;
+			if let Err(err) = socket.set_nodelay(true) {
+				log::warn!("set nodelay failed: {:?}", err);
+			}
+			match tls_connector {
+				None => Ok(TlsOrPlain::Plain(socket)),
+				Some(connector) => {
+					let dns_name = DNSNameRef::try_from_ascii_str(host)?;
+					let tls_stream = connector.connect(dns_name, socket).await?;
+					Ok(TlsOrPlain::Tls(tls_stream))
+				}
+			}
+		}
+		_ = timeout => Err(WsHandshakeError::Timeout(timeout_dur))
 	}
 }
 
@@ -251,8 +343,8 @@ impl From<io::Error> for WsHandshakeError {
 	}
 }
 
-impl From<crate::tokio::InvalidDNSNameError> for WsHandshakeError {
-	fn from(err: crate::tokio::InvalidDNSNameError) -> WsHandshakeError {
+impl From<InvalidDNSNameError> for WsHandshakeError {
+	fn from(err: InvalidDNSNameError) -> WsHandshakeError {
 		WsHandshakeError::InvalidDnsName(err)
 	}
 }
@@ -284,34 +376,32 @@ pub struct Target {
 	path_and_query: String,
 }
 
-impl Target {
-	/// Parse an URL String to a WebSocket address.
-	pub fn parse(url: impl AsRef<str>) -> Result<Self, WsHandshakeError> {
-		let url =
-			url::Url::parse(url.as_ref()).map_err(|e| WsHandshakeError::Url(format!("Invalid URL: {}", e).into()))?;
-		let mode = match url.scheme() {
-			"ws" => Mode::Plain,
-			"wss" => Mode::Tls,
+impl TryFrom<Uri> for Target {
+	type Error = WsHandshakeError;
+
+	fn try_from(uri: Uri) -> Result<Self, Self::Error> {
+		let mode = match uri.scheme_str() {
+			Some("ws") => Mode::Plain,
+			Some("wss") => Mode::Tls,
 			_ => return Err(WsHandshakeError::Url("URL scheme not supported, expects 'ws' or 'wss'".into())),
 		};
-		let host =
-			url.host_str().map(ToOwned::to_owned).ok_or_else(|| WsHandshakeError::Url("No host in URL".into()))?;
-		let port = url.port_or_known_default().ok_or_else(|| WsHandshakeError::Url("No port number in URL".into()))?;
+		let host = uri.host().map(ToOwned::to_owned).ok_or_else(|| WsHandshakeError::Url("No host in URL".into()))?;
+		let port = uri
+			.port_u16()
+			.ok_or_else(|| WsHandshakeError::Url("No port number in URL (default port is not supported)".into()))?;
 		let host_header = format!("{}:{}", host, port);
-		let mut path_and_query = url.path().to_owned();
-		if let Some(query) = url.query() {
-			path_and_query.push('?');
-			path_and_query.push_str(query);
-		}
-		// NOTE: `Url::socket_addrs` is using the default port if it's missing (ws:// - 80, wss:// - 443)
-		let sockaddrs = url.socket_addrs(|| None).map_err(WsHandshakeError::ResolutionFailed)?;
-		Ok(Self { sockaddrs, host, host_header, mode, path_and_query })
+		let parts = uri.into_parts();
+		let path_and_query = parts.path_and_query.ok_or_else(|| WsHandshakeError::Url("No path in URL".into()))?;
+		let sockaddrs = host_header.to_socket_addrs().map_err(WsHandshakeError::ResolutionFailed)?;
+		Ok(Self { sockaddrs: sockaddrs.collect(), host, host_header, mode, path_and_query: path_and_query.to_string() })
 	}
 }
 
 #[cfg(test)]
 mod tests {
-	use super::{Mode, Target, WsHandshakeError};
+	use super::{Mode, Target, Uri, WsHandshakeError};
+	use http::uri::InvalidUri;
+	use std::convert::TryInto;
 
 	fn assert_ws_target(target: Target, host: &str, host_header: &str, mode: Mode, path_and_query: &str) {
 		assert_eq!(&target.host, host);
@@ -320,53 +410,51 @@ mod tests {
 		assert_eq!(&target.path_and_query, path_and_query);
 	}
 
+	fn parse_target(uri: &str) -> Result<Target, WsHandshakeError> {
+		uri.parse::<Uri>().map_err(|e: InvalidUri| WsHandshakeError::Url(e.to_string().into()))?.try_into()
+	}
+
 	#[test]
 	fn ws_works() {
-		let target = Target::parse("ws://127.0.0.1:9933").unwrap();
+		let target = parse_target("ws://127.0.0.1:9933").unwrap();
 		assert_ws_target(target, "127.0.0.1", "127.0.0.1:9933", Mode::Plain, "/");
 	}
 
 	#[test]
 	fn wss_works() {
-		let target = Target::parse("wss://kusama-rpc.polkadot.io:443").unwrap();
+		let target = parse_target("wss://kusama-rpc.polkadot.io:443").unwrap();
 		assert_ws_target(target, "kusama-rpc.polkadot.io", "kusama-rpc.polkadot.io:443", Mode::Tls, "/");
 	}
 
 	#[test]
 	fn faulty_url_scheme() {
-		let err = Target::parse("http://kusama-rpc.polkadot.io:443").unwrap_err();
+		let err = parse_target("http://kusama-rpc.polkadot.io:443").unwrap_err();
 		assert!(matches!(err, WsHandshakeError::Url(_)));
 	}
 
 	#[test]
 	fn faulty_port() {
-		let err = Target::parse("ws://127.0.0.1:-43").unwrap_err();
+		let err = parse_target("ws://127.0.0.1:-43").unwrap_err();
 		assert!(matches!(err, WsHandshakeError::Url(_)));
-		let err = Target::parse("ws://127.0.0.1:99999").unwrap_err();
+		let err = parse_target("ws://127.0.0.1:99999").unwrap_err();
 		assert!(matches!(err, WsHandshakeError::Url(_)));
 	}
 
-	#[test]
-	fn default_port_works() {
-		let target = Target::parse("ws://127.0.0.1").unwrap();
-		assert_ws_target(target, "127.0.0.1", "127.0.0.1:80", Mode::Plain, "/");
-	}
-
 	#[test]
 	fn url_with_path_works() {
-		let target = Target::parse("wss://127.0.0.1/my-special-path").unwrap();
+		let target = parse_target("wss://127.0.0.1:443/my-special-path").unwrap();
 		assert_ws_target(target, "127.0.0.1", "127.0.0.1:443", Mode::Tls, "/my-special-path");
 	}
 
 	#[test]
 	fn url_with_query_works() {
-		let target = Target::parse("wss://127.0.0.1/my?name1=value1&name2=value2").unwrap();
+		let target = parse_target("wss://127.0.0.1:443/my?name1=value1&name2=value2").unwrap();
 		assert_ws_target(target, "127.0.0.1", "127.0.0.1:443", Mode::Tls, "/my?name1=value1&name2=value2");
 	}
 
 	#[test]
 	fn url_with_fragment_is_ignored() {
-		let target = Target::parse("wss://127.0.0.1/my.htm#ignore").unwrap();
+		let target = parse_target("wss://127.0.0.1:443/my.htm#ignore").unwrap();
 		assert_ws_target(target, "127.0.0.1", "127.0.0.1:443", Mode::Tls, "/my.htm");
 	}
 }
